@@ -15,14 +15,46 @@ import operator
 from random import randrange, sample, choice
 
 from six import PY2
-from six.moves import filterfalse, map, range, zip, zip_longest
+from six.moves import filter, filterfalse, map, range, zip, zip_longest
 
-__all__ = ['take', 'tabulate', 'consume', 'nth', 'quantify', 'padnone',
-           'ncycles', 'dotproduct', 'flatten', 'repeatfunc', 'pairwise',
-           'grouper', 'roundrobin', 'powerset', 'unique_everseen',
-           'unique_justseen', 'iter_except', 'random_product',
-           'random_permutation', 'random_combination',
-           'random_combination_with_replacement']
+__all__ = [
+    'accumulate', 'all_equal', 'consume', 'dotproduct', 'first_true',
+    'flatten', 'grouper', 'iter_except', 'ncycles', 'nth', 'padnone',
+    'pairwise', 'partition', 'powerset', 'quantify',
+    'random_combination_with_replacement', 'random_combination',
+    'random_permutation', 'random_product', 'repeatfunc', 'roundrobin',
+    'tabulate', 'tail', 'take', 'unique_everseen', 'unique_justseen'
+]
+
+
+def accumulate(iterable, func=operator.add):
+    """
+    Return an iterator whose items are the accumulated results of a function
+    (specified by the optional *func* argument) that takes two arguments.
+    By default, returns accumulated sums with ``operator.add()``.
+
+        >>> list(accumulate([1, 2, 3, 4, 5]))  # Running sum
+        [1, 3, 6, 10, 15]
+        >>> list(accumulate([1, 2, 3], func=operator.mul))  # Running product
+        [1, 2, 6]
+        >>> list(accumulate([0, 1, -1, 2, 3, 2], func=max))  # Running maximum
+        [0, 1, 1, 2, 3, 3]
+
+    This function is available in the ``itertools`` module for Python 3.2 and
+    greater.
+
+    """
+    it = iter(iterable)
+    try:
+        total = next(it)
+    except StopIteration:
+        return
+    else:
+        yield total
+
+    for element in it:
+        total = func(total, element)
+        yield total
 
 
 def take(n, iterable):
@@ -54,6 +86,18 @@ def tabulate(function, start=0):
     return map(function, count(start))
 
 
+def tail(n, iterable):
+    """
+    Return an iterator over the last n items"
+
+        >>> t = tail(3, 'ABCDEFG')
+        >>> list(t)
+        ['E', 'F', 'G']
+
+    """
+    return iter(deque(iterable, maxlen=n))
+
+
 def consume(iterator, n=None):
     """Advance the iterator n-steps ahead. If n is none, consume entirely.
 
@@ -106,6 +150,20 @@ def nth(iterable, n, default=None):
     return next(islice(iterable, n, None), default)
 
 
+def all_equal(iterable):
+    """
+    Returns True if all the elements are equal to each other.
+
+        >>> all_equal('aaaa')
+        True
+        >>> all_equal('aaab')
+        False
+
+    """
+    g = groupby(iterable)
+    return next(g, True) and not next(g, False)
+
+
 def quantify(iterable, pred=bool):
     """Return the how many times the predicate is true
 
@@ -217,6 +275,24 @@ def roundrobin(*iterables):
             nexts = cycle(islice(nexts, pending))
 
 
+def partition(pred, iterable):
+    """
+    Returns a 2-tuple of iterables derived from the input iterable.
+    The first yields the items that have ``pred(item) == False``.
+    The first yields the items that have ``pred(item) == False``.
+
+        >>> is_odd = lambda x: x % 2 != 0
+        >>> iterable = range(10)
+        >>> even_items, odd_items = partition(is_odd, iterable)
+        >>> list(even_items), list(odd_items)
+        ([0, 2, 4, 6, 8], [1, 3, 5, 7, 9])
+
+    """
+    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
+    t1, t2 = tee(iterable)
+    return filterfalse(pred, t1), filter(pred, t2)
+
+
 def powerset(iterable):
     """Yields all possible subsets of the iterable
 
@@ -229,26 +305,42 @@ def powerset(iterable):
 
 
 def unique_everseen(iterable, key=None):
-    """Yield unique elements, preserving order.
-
+    """
+    Yield unique elements, preserving order.
         >>> list(unique_everseen('AAAABBBCCDAABBB'))
         ['A', 'B', 'C', 'D']
         >>> list(unique_everseen('ABBCcAD', str.lower))
         ['A', 'B', 'C', 'D']
 
+    Sequences with a mix of hashable and unhashable items can be used.
+    The function will be slower (i.e., O(N^2)) for unhashable items.
+
     """
-    seen = set()
-    seen_add = seen.add
+    seenset = set()
+    seenset_add = seenset.add
+    seenlist = []
+    seenlist_add = seenlist.append
     if key is None:
-        for element in filterfalse(seen.__contains__, iterable):
-            seen_add(element)
-            yield element
+        for element in iterable:
+            try:
+                if element not in seenset:
+                    seenset_add(element)
+                    yield element
+            except TypeError as e:
+                if element not in seenlist:
+                    seenlist_add(element)
+                    yield element
     else:
         for element in iterable:
             k = key(element)
-            if k not in seen:
-                seen_add(k)
-                yield element
+            try:
+                if k not in seenset:
+                    seenset_add(k)
+                    yield element
+            except TypeError as e:
+                if k not in seenlist:
+                    seenlist_add(k)
+                    yield element
 
 
 def unique_justseen(iterable, key=None):
@@ -284,6 +376,26 @@ def iter_except(func, exception, first=None):
         pass
 
 
+def first_true(iterable, default=False, pred=None):
+    """
+    Returns the first true value in the iterable.
+
+    If no true value is found, returns *default*
+
+    If *pred* is not None, returns the first item for which
+    ``pred(item) == True`` .
+
+        >>> first_true(range(10))
+        1
+        >>> first_true(range(10), pred=lambda x: x > 5)
+        6
+        >>> first_true(range(10), default='missing', pred=lambda x: x > 9)
+        'missing'
+
+    """
+    return next(filter(pred, iterable), default)
+
+
 def random_product(*args, **kwds):
     """Returns a random pairing of items from each iterable argument
 
